SET SERVEROUTPUT ON

--
-- Drop Table procedure
CREATE OR REPLACE PROCEDURE drop_if_exists(table_name VARCHAR2)
IS
  table_inexistent EXCEPTION;                  -- table does not exists
  PRAGMA EXCEPTION_INIT(table_inexistent,-942);-- associate with error ORA-00942
BEGIN 
  EXECUTE IMMEDIATE ('DROP TABLE '|| table_name || ' CASCADE CONSTRAINTS');
EXCEPTION
WHEN table_inexistent THEN
  RETURN;
END drop_if_exists;
/

--
-- Drop column procedure
CREATE OR REPLACE PROCEDURE drop_column_if_exists(table_name  VARCHAR2
                                                 ,column_name VARCHAR2)
IS
  column_inexistent EXCEPTION;                  -- column does not exist
  PRAGMA EXCEPTION_INIT(column_inexistent,-904);-- associate with error ORA-00904
BEGIN
  EXECUTE IMMEDIATE ('ALTER TABLE '|| table_name || ' DROP (' || column_name || ')');
EXCEPTION
WHEN column_inexistent THEN
  RETURN;
END drop_column_if_exists;
/

--
-- Rename column procedure
CREATE OR REPLACE PROCEDURE rename_column(table_name  VARCHAR2
                                         ,column_name VARCHAR2
                                         ,column_type VARCHAR2)
IS
BEGIN
  drop_column_if_exists(table_name, column_name || '_tmp');
  EXECUTE IMMEDIATE ('ALTER TABLE ' || table_name || ' ADD (' || column_name || '_tmp ' || column_type || ')');
  EXECUTE IMMEDIATE ('UPDATE ' || table_name || ' SET ' || column_name || '_tmp=' || column_name);
  EXECUTE IMMEDIATE ('COMMIT');
  EXECUTE IMMEDIATE ('ALTER TABLE ' || table_name || ' DROP (' || column_name || ')');
  EXECUTE IMMEDIATE ('ALTER TABLE ' || table_name || ' RENAME COLUMN ' || column_name || '_tmp TO ' || column_name);
END rename_column;
/

--
-- Main
DECLARE
  v_major INTEGER := 0;
  v_minor INTEGER := 0;
  v_patch INTEGER := 0;
  
  v_req_major     CONSTANT INTEGER  := 2;
  v_req_minor     CONSTANT INTEGER  := 0;
  v_req_patch_min CONSTANT INTEGER  := 0;
  v_req_patch_max CONSTANT INTEGER  := 1;
  
  v_next_major CONSTANT INTEGER := 2;
  v_next_minor CONSTANT INTEGER := 1;
  v_next_patch CONSTANT INTEGER := 0;
 
BEGIN

  DBMS_OUTPUT.PUT_LINE('Applying Patch...');
  -- Get schema version
  SELECT MAJOR,MINOR,PATCH INTO v_major,v_minor,v_patch FROM T_SCHEMA_VERS WHERE ROWNUM <=1;
  -- check version
  IF v_major = v_req_major AND  v_minor = v_req_minor AND (v_patch BETWEEN  v_req_patch_min AND v_req_patch_max) THEN
     -- apply patch

     -- Mofify T_CHANNEL Table
     rename_column('T_CHANNEL','LAST_ACTIVE','TIMESTAMP WITH TIME ZONE');
     -- Modify T_JOB Table
     rename_column('T_JOB','SUBMIT_TIME','TIMESTAMP WITH TIME ZONE DEFAULT SYSTIMESTAMP AT TIME ZONE ''+00:00''');
     -- Modify T_TRANSFER Table
     rename_column('T_TRANSFER','TRANSFER_TIME','TIMESTAMP WITH TIME ZONE');
     -- Modify T_TRANCE Table
     rename_column('T_TRACE','TRACE_TIMESTAMP','TIMESTAMP WITH TIME ZONE DEFAULT SYSTIMESTAMP AT TIME ZONE ''+00:00''');
     -- Modify T_TRACE AUTO INC Trigger
     EXECUTE IMMEDIATE ('CREATE OR REPLACE TRIGGER trace_trace_id_auto_inc
                         BEFORE INSERT ON t_trace FOR EACH ROW WHEN (new.trace_id IS NULL) BEGIN
                         SELECT trace_trace_id_seq.nextval INTO :new.trace_id from dual;END;');
     -- Add T_AGENT Table
     drop_if_exists('t_agent');
     EXECUTE IMMEDIATE ('CREATE TABLE t_agent (
       agent_name           VARCHAR2(255)
                            CONSTRAINT agent_agent_name_not_null NOT NULL
                            CONSTRAINT agent_agent_name_pk PRIMARY KEY
       ,agent_dn            VARCHAR2(1024)
       ,agent_host          VARCHAR2(255)
       ,agent_type          VARCHAR2(8)
       ,last_active         TIMESTAMP WITH TIME ZONE
       ,agent_state         VARCHAR2(16)
       ,agent_version       VARCHAR2(16)
       ,contact             VARCHAR2(255)
       )');
     -- Add T_TRANSFER_CUMULATIVE_STATS Table
     drop_if_exists('t_transfer_cumulative_stats');
     EXECUTE IMMEDIATE ('CREATE TABLE t_transfer_cumulative_stats (
        time                TIMESTAMP
       ,channel_name        VARCHAR2(32)
       ,vo_name             VARCHAR2(50)
       ,num_jobs_submitted  NUMBER(5) DEFAULT 0
       ,num_jobs_pending    NUMBER(5) DEFAULT 0
       ,num_jobs_active     NUMBER(5) DEFAULT 0
       ,num_jobs_done       NUMBER(5) DEFAULT 0
       ,num_jobs_hold       NUMBER(5) DEFAULT 0
       ,num_jobs_failed     NUMBER(5) DEFAULT 0
       ,num_jobs_canceling  NUMBER(5) DEFAULT 0
       ,num_jobs_canceled   NUMBER(5) DEFAULT 0
       ,num_jobs_finished   NUMBER(5) DEFAULT 0
       ,num_jobs_finished_dirty  NUMBER(5) DEFAULT 0
       ,num_files_submitted NUMBER(5) DEFAULT 0
       ,num_files_pending   NUMBER(5) DEFAULT 0
       ,num_files_active    NUMBER(5) DEFAULT 0
       ,num_files_done      NUMBER(5) DEFAULT 0
       ,num_files_failed    NUMBER(5) DEFAULT 0
       ,num_files_waiting   NUMBER(5) DEFAULT 0
       ,num_files_hold      NUMBER(5) DEFAULT 0
       ,num_files_canceling NUMBER(5) DEFAULT 0
       ,num_files_canceled  NUMBER(5) DEFAULT 0
       ,num_files_finished  NUMBER(5) DEFAULT 0
       ,num_files_catalog_failed NUMBER(5) DEFAULT 0
       ,CONSTRAINT transfer_cumulative_stats_pk PRIMARY KEY (time, channel_name, vo_name) 
       )');
     -- Add T_TRANSFER_PERIOD_STATS Table
     drop_if_exists('t_transfer_period_stats');
     EXECUTE IMMEDIATE ('CREATE TABLE t_transfer_period_stats (
       start_time	                TIMESTAMP
       ,end_time	                TIMESTAMP
       ,channel_name        VARCHAR2(32)
       ,vo_name             VARCHAR2(50)
       ,average_rate FLOAT
       ,bytes_written INTEGER
       ,CONSTRAINT transfer_stats_period_pk PRIMARY KEY (start_time, end_time, channel_name, vo_name)
       )');

     -- Add T_AGENT INDEXES
     EXECUTE IMMEDIATE ('CREATE INDEX agent_agent_state ON t_agent(agent_state)');
     -- Add T_TRANSFER_CUMULATIVE_STATS INDEXES
     EXECUTE IMMEDIATE ('CREATE INDEX transfer_cum_stats_time    ON t_transfer_cumulative_stats(time)'); 
     EXECUTE IMMEDIATE ('CREATE INDEX transfer_cum_stats_ch_name ON t_transfer_cumulative_stats(channel_name)'); 
     EXECUTE IMMEDIATE ('CREATE INDEX transfer_cum_stats_vo_name ON t_transfer_cumulative_stats(vo_name)'); 
     -- Add T_TRANSFER_PERIOD_STATS INDEXES
     EXECUTE IMMEDIATE ('CREATE INDEX transfer_per_stats_st_time ON t_transfer_period_stats(start_time)');
     EXECUTE IMMEDIATE ('CREATE INDEX transfer_per_stats_en_time ON t_transfer_period_stats(end_time)');
     EXECUTE IMMEDIATE ('CREATE INDEX transfer_per_stats_ch_name ON t_transfer_period_stats(channel_name)');
     EXECUTE IMMEDIATE ('CREATE INDEX transfer_per_stats_vo_name ON t_transfer_period_stats(vo_name)');

     -- Add T_CHANNEL unique source/dest pair constraint
     EXECUTE IMMEDIATE ('ALTER TABLE t_channel ADD CONSTRAINT channel_source_dest_uq UNIQUE (source_site, dest_site)');

     -- Update Schema Version
     UPDATE T_SCHEMA_VERS SET MAJOR=v_next_major, MINOR=v_next_minor, PATCH=v_next_patch; 
     DBMS_OUTPUT.PUT_LINE('SUCCESS: Patch applied. New version is ' || TO_CHAR(v_next_major) || '.' || TO_CHAR(v_next_minor) || '.' || TO_CHAR(v_next_patch) );
  ELSIF v_major = v_next_major AND  v_minor = v_next_minor AND v_patch = v_next_patch THEN
     DBMS_OUTPUT.PUT_LINE('SUCCESS: Patch already applied. Version is ' || TO_CHAR(v_next_major) || '.' || TO_CHAR(v_next_minor) || '.' || TO_CHAR(v_next_patch) );
  ELSE
     DBMS_OUTPUT.PUT_LINE('ERROR: Cannot apply patch. Version ' || TO_CHAR(v_major) || '.' || TO_CHAR(v_minor) || '.' || TO_CHAR(v_patch) || ' does not match the required one (' || TO_CHAR(v_req_major) || '.' || TO_CHAR(v_req_minor) || '.' || TO_CHAR(v_req_patch_min) || '-' ||  TO_CHAR(v_req_patch_max) || ')');
  END IF;
END;
/
DROP PROCEDURE drop_if_exists;
DROP PROCEDURE drop_column_if_exists;
DROP PROCEDURE rename_column;

