SET SERVEROUTPUT ON

--
-- Add index if it does not exist
CREATE OR REPLACE PROCEDURE add_index(index_name  VARCHAR2
                   ,index_stmt  VARCHAR2)
IS
  index_exists EXCEPTION;                  --index already exists
  PRAGMA EXCEPTION_INIT(index_exists,-955); -- associate with error ORA-00955
BEGIN
  EXECUTE IMMEDIATE ('CREATE INDEX ' || index_name || ' ON ' || index_stmt || ' ONLINE');
  DBMS_OUTPUT.PUT_LINE('INFO: Index [' || index_name || '] added');
EXCEPTION
WHEN index_exists THEN
  DBMS_OUTPUT.PUT_LINE('INFO: Index [' || index_name || '] already exists. Skip it');
  RETURN;
END; --add_index
/

--
-- Drop Table procedure
CREATE OR REPLACE PROCEDURE drop_if_exists(table_name VARCHAR2)
IS
  table_inexistent EXCEPTION;                   -- table does not exists
  PRAGMA EXCEPTION_INIT(table_inexistent,-942); -- associate with error ORA-00942
BEGIN 
  EXECUTE IMMEDIATE ('DROP TABLE '|| table_name || ' CASCADE CONSTRAINTS');
EXCEPTION
WHEN table_inexistent THEN
  RETURN;
END drop_if_exists;
/

--
-- Drop Sequence procedure
CREATE OR REPLACE PROCEDURE drop_seq_if_exists(seq_name VARCHAR2)
IS
  seq_inexistent EXCEPTION;                  -- seq does not exists
  PRAGMA EXCEPTION_INIT(seq_inexistent,-2289);-- associate with error ORA-02289
BEGIN 
  EXECUTE IMMEDIATE ('DROP SEQUENCE '|| seq_name);
EXCEPTION
WHEN seq_inexistent THEN
  RETURN;
END drop_seq_if_exists;
/


--
-- Main
DECLARE
  v_major INTEGER := 0;
  v_minor INTEGER := 0;
  v_patch INTEGER := 0;
  
  v_req_major     CONSTANT INTEGER  := 2;
  v_req_minor     CONSTANT INTEGER  := 2;
  v_req_patch_min CONSTANT INTEGER  := 1;
  v_req_patch_max CONSTANT INTEGER  := 1;
  
  v_next_major CONSTANT INTEGER := 3;
  v_next_minor CONSTANT INTEGER := 0;
  v_next_patch CONSTANT INTEGER := 0;

BEGIN

  DBMS_OUTPUT.PUT_LINE('Applying Patch...');
  -- Get schema version
  SELECT MAJOR,MINOR,PATCH INTO v_major,v_minor,v_patch FROM T_SCHEMA_VERS WHERE ROWNUM <=1;
  -- check version
  IF v_major = v_req_major AND  v_minor = v_req_minor AND (v_patch BETWEEN  v_req_patch_min AND v_req_patch_max) THEN
      -- apply patch

      -- update t_channel table
      DBMS_OUTPUT.PUT_LINE('Updating T_CHANNEL...');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL ADD TCP_BUFFER_SIZE VARCHAR2(24)');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL ADD MESSAGE VARCHAR2(1024)');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL ADD LAST_MODIFICATION TIMESTAMP WITH TIME ZONE');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL ADD ADMIN_DN VARCHAR2(1024)');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL ADD SE_LIMIT INTEGER DEFAULT NULL');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL ADD CHANNEL_TYPE VARCHAR2(20)');
      
      -- add t_audit_channel table
      DBMS_OUTPUT.PUT_LINE('Adding T_AUDIT_CHANNEL...');
      drop_if_exists('T_AUDIT_CHANNEL');
      EXECUTE IMMEDIATE('CREATE TABLE t_audit_channel (
         audit_id		INTEGER  CONSTRAINT audit_channel_id_pk PRIMARY KEY
         ,channel_name     	VARCHAR2(32)
                              CONSTRAINT audit_channel_ch_name_not_null NOT NULL
         ,contact         	   VARCHAR2(255) NULL
         ,bandwidth       	   NUMBER NULL
         ,nostreams       	   INTEGER NULL
         ,nofiles             INTEGER NULL
         ,tcp_buffer_size     VARCHAR2(24) NULL
         ,nominal_throughput	NUMBER NULL
         ,channel_state	      VARCHAR2(30) NULL
         ,message             VARCHAR2(1024) NULL
         ,modification_time   TIMESTAMP WITH TIME ZONE
         ,admin_dn            VARCHAR2(1024)
         ,fts_node		      VARCHAR2(1024)
         ,ip_addr		         VARCHAR2(255)
         )');
      EXECUTE IMMEDIATE('CREATE SEQUENCE audit_chan_id_seq');
      EXECUTE IMMEDIATE('CREATE OR REPLACE TRIGGER audit_channel_id_auto_inc
                           BEFORE INSERT ON t_audit_channel
                           FOR EACH ROW
                           WHEN (new.audit_id IS NULL)
                           BEGIN
                              SELECT audit_chan_id_seq.nextval
                              INTO   :new.audit_id from dual;
                           END;');
      
      -- add t_bad_ses table, sequence, trigger and index
      DBMS_OUTPUT.PUT_LINE('Adding T_BAD_SES...');
      drop_if_exists('T_BAD_SES');
      EXECUTE IMMEDIATE('CREATE TABLE t_bad_ses (
            se_id                INTEGER
                                 CONSTRAINT bad_ses_id_pk PRIMARY KEY   
            ,se_hostname         VARCHAR2(256)
            ,message             VARCHAR2(1024) DEFAULT NULL
            ,addition_time       TIMESTAMP WITH TIME ZONE
            ,admin_dn            VARCHAR2(1024) )');
      EXECUTE IMMEDIATE('CREATE SEQUENCE se_bad_id_seq');
      EXECUTE IMMEDIATE('CREATE OR REPLACE TRIGGER bad_se_id_auto_inc
                        BEFORE INSERT ON t_bad_ses
                        FOR EACH ROW
                        WHEN (new.se_id IS NULL)
                        BEGIN
                          SELECT se_bad_id_seq.nextval
                          INTO   :new.se_id from dual;
                        END;');
      add_index('bad_se_hostname', 't_bad_ses(se_hostname)');


      -- add t_transfer_extra
      DBMS_OUTPUT.PUT_LINE('Adding T_TRANSFER_EXTRA...');
      drop_if_exists('T_TRANSFER_EXTRA');
      EXECUTE IMMEDIATE('CREATE TABLE t_transfer_extra (
                        message_id	         INTEGER
                                             CONSTRAINT extra_message_id_pk PRIMARY KEY
                        ,request_id          VARCHAR2(255)
                        ,file_id              INTEGER
                        ,timestamp           TIMESTAMP WITH TIME ZONE DEFAULT SYSTIMESTAMP AT TIME ZONE ''+00:00''
                        ,debug_level         VARCHAR2(20) DEFAULT NULL
                        ,scope               VARCHAR2(20) DEFAULT NULL 
                        ,phase               VARCHAR(20) DEFAULT NULL
                        ,category            VARCHAR(64) DEFAULT NULL
                        ,message             VARCHAR(1024) DEFAULT NULL )');
      EXECUTE IMMEDIATE('CREATE SEQUENCE transfer_extra_id_seq');
      EXECUTE IMMEDIATE('CREATE OR REPLACE TRIGGER transfer_ex_id_auto_inc
                        BEFORE INSERT ON t_transfer_extra
                        FOR EACH ROW
                        WHEN (new.message_id IS NULL)
                        BEGIN
                          SELECT transfer_extra_id_seq.nextval
                          INTO   :new.message_id from dual;
                        END;');
      add_index('extra_reqid_fileid', 't_transfer_extra(request_id,file_id)'); 

      -- update t_channel_share
      DBMS_OUTPUT.PUT_LINE('Updating T_CHANNEL_VO_SHARE...');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL_VO_SHARE ADD CHANNEL_LIMIT INTEGER DEFAULT 0');
      EXECUTE IMMEDIATE('ALTER TABLE T_CHANNEL_VO_SHARE ADD CHANNEL_VOSTATE VARCHAR2(10) DEFAULT ''ON''');

      -- alter t_job
      DBMS_OUTPUT.PUT_LINE('Updating T_JOB...');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB MODIFY JOB_STATE VARCHAR2(32)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD CANCEL_JOB CHAR(1)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD CRED_ID VARCHAR2(100)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD FINISH_TIME TIMESTAMP WITH TIME ZONE');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB MODIFY SUBMIT_HOST VARCHAR2(255)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD SPACE_TOKEN VARCHAR2(255)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD STORAGE_CLASS VARCHAR2(255)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD MYPROXY_SERVER VARCHAR2(255)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD SRC_CATALOG VARCHAR2(1024)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD SRC_CATALOG_TYPE VARCHAR2(1024)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD DEST_CATALOG VARCHAR2(1024)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD DEST_CATALOG_TYPE VARCHAR2(1024)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD SOURCE_SE VARCHAR2(255)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD DEST_SE VARCHAR2(255)');
      EXECUTE IMMEDIATE('ALTER TABLE T_JOB ADD OVERWRITE_FLAG CHAR(1)');

      -- alter t_file
      DBMS_OUTPUT.PUT_LINE('Updating T_FILE...');
      EXECUTE IMMEDIATE('DROP INDEX IDX_REPORT_FILE');
      EXECUTE IMMEDIATE('ALTER TABLE T_FILE MODIFY FILE_STATE VARCHAR2(32)');
      EXECUTE IMMEDIATE('CREATE INDEX idx_report_file ON t_file (finish_time,file_state,file_id,job_id,filesize) ONLINE');
      EXECUTE IMMEDIATE('ALTER TABLE T_FILE ADD CATALOG_FAILURES INTEGER');
      EXECUTE IMMEDIATE('ALTER TABLE T_FILE ADD PRESTAGE_FAILURES INTEGER');
      EXECUTE IMMEDIATE('ALTER TABLE T_FILE ADD INTERNAL_FILE_PARAMS VARCHAR2(255)');
       
      -- alter t_transfer
      DBMS_OUTPUT.PUT_LINE('Updating T_TRANSFER...');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD JOB_ID CHAR(36)
                           REFERENCES T_JOB(JOB_ID)');
      EXECUTE IMMEDIATE('DROP INDEX IDX_REPORT_TRANSFER');
      EXECUTE IMMEDIATE('DROP INDEX TRANSFER_KEY_SELECT');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER MODIFY TRANSFER_STATE VARCHAR2(32)');
      EXECUTE IMMEDIATE('CREATE index transfer_key_select ON t_transfer(transfer_state,file_id,request_id,transfer_id,transfer_time)');
      EXECUTE IMMEDIATE('CREATE index idx_report_transfer ON t_transfer (transfer_state,reason_class,file_id,duration,bytes_written)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD PREPARE_TIME TIMESTAMP WITH TIME ZONE');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD SRC_PREP_DURATION NUMBER(12,3)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD DEST_PREP_DURATION NUMBER(12,3)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD TX_DURATION NUMBER(12,3)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD SRC_FINAL_DURATION NUMBER(12,3)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD DEST_FINAL_DURATION NUMBER(12,3)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD THROUGHPUT NUMBER');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD FILESIZE INTEGER');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD TRANSFER_TYPE VARCHAR(32)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD SOURCE_SRM_TOKEN VARCHAR(256)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD DEST_SRM_TOKEN VARCHAR(256)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD SOURCE_HOST VARCHAR(256)');
      EXECUTE IMMEDIATE('ALTER TABLE T_TRANSFER ADD DEST_HOST VARCHAR(256)');


      -- add t_stage_req table
      DBMS_OUTPUT.PUT_LINE('Updating T_STAGE_REQ...');
      drop_if_exists('T_STAGE_REQ');
      EXECUTE IMMEDIATE('CREATE TABLE t_stage_req (
           request_id            VARCHAR2(255)
                                 CONSTRAINT stagereq_req_id_not_null NOT NULL
           ,file_id              INTEGER
                                 CONSTRAINT stagereq_file_id_not_null NOT NULL
                                 REFERENCES t_file(file_id)
           ,stage_id             INTEGER
                                 CONSTRAINT stagereq_stage_id_not_null NOT NULL
           ,job_id               CHAR(36)
                                 CONSTRAINT stagereq_job_id_not_null NOT NULL
                                 REFERENCES t_job(job_id)
           ,stage_state          VARCHAR2(32)
                                 CONSTRAINT stagereq_state_not_null NOT NULL
           ,turl                 VARCHAR2(1100)
           ,request_time         TIMESTAMP WITH TIME ZONE
           ,duration             NUMBER(12,2)
           ,final_duration       NUMBER(12,3)
           ,reason_class         VARCHAR2(32)
           ,reason               VARCHAR2(1024)
           ,filesize         	   INTEGER
           ,CONSTRAINT stagereq_pk PRIMARY KEY (request_id, file_id)
         )');


      -- drop the unused credential tables if they are already there (the new delegation has a different schema)
      DBMS_OUTPUT.PUT_LINE('Dropping old T_CREDENTIAL and T_CREDENTIAL_CACHE...');
      drop_if_exists('T_CREDENTIAL');
      drop_if_exists('T_CREDENTIAL_CACHE');

      -- drop unused stats tables
      drop_if_exists('T_TRANSFER_CUMULATIVE_STATS');
      drop_if_exists('T_TRANSFER_PERIOD_STATS');

       -- drop unused t_trace table and its sequence
      drop_if_exists('T_TRACE');
      drop_seq_if_exists('TRACE_TRACE_ID_SEQ');

      -- add new indices
      DBMS_OUTPUT.PUT_LINE('Adding T_AUDIT_CHANNEL indices...');
      add_index('channel_audit_channel_name', 't_audit_channel(channel_name)');
      add_index('channel_audit_modif_time', 't_audit_channel(modification_time)');
      add_index('channel_audit_channel_state', 't_audit_channel(channel_state)');

      DBMS_OUTPUT.PUT_LINE('Adding T_JOB indices...');
      add_index('job_cred_id', 't_job(user_dn, cred_id)');

      DBMS_OUTPUT.PUT_LINE('Adding T_TRANSFER indices...');
      add_index('transfer_job_id', 't_transfer(job_id)');

      DBMS_OUTPUT.PUT_LINE('Adding T_STAGE_REQ indices...');
      add_index('stagereq_request_id', 't_stage_req(request_id)');
      add_index('stagereq_file_id', 't_stage_req(file_id)');
      add_index('stagereq_job_id', 't_stage_req(job_id)');
      add_index('stagereq_stage_state', 't_stage_req(stage_state)');

      -- Update Schema Version
      DBMS_OUTPUT.PUT_LINE('Updating schema version...');
      UPDATE T_SCHEMA_VERS SET MAJOR=v_next_major, MINOR=v_next_minor, PATCH=v_next_patch; 
      DBMS_OUTPUT.PUT_LINE('SUCCESS: Patch applied. New version is ' || TO_CHAR(v_next_major) || '.' || TO_CHAR(v_next_minor) || '.' || TO_CHAR(v_next_patch) );
  ELSIF v_major = v_next_major AND  v_minor = v_next_minor AND v_patch = v_next_patch THEN
     DBMS_OUTPUT.PUT_LINE('SUCCESS: Patch already applied. Version is ' || TO_CHAR(v_next_major) || '.' || TO_CHAR(v_next_minor) || '.' || TO_CHAR(v_next_patch) );
  ELSE
     DBMS_OUTPUT.PUT_LINE('ERROR: Cannot apply patch. Version ' || TO_CHAR(v_major) || '.' || TO_CHAR(v_minor) || '.' || TO_CHAR(v_patch) || ' does not match the required one (' || TO_CHAR(v_req_major) || '.' || TO_CHAR(v_req_minor) || '.' || TO_CHAR(v_req_patch_min) || '-' ||  TO_CHAR(v_req_patch_max) || ')');
  END IF;
END;
/
