SET SERVEROUTPUT ON SIZE 1000000

DECLARE
  
  v_major INTEGER := 0;
  v_minor INTEGER := 0;
  v_patch INTEGER := 0;
  v_upgrade_state VARCHAR2(24);
  
  v_req_major     CONSTANT INTEGER  := 3;
  v_req_minor     CONSTANT INTEGER  := 0;
  v_req_patch_min CONSTANT INTEGER  := 0;
  v_req_patch_max CONSTANT INTEGER  := 0;
  
  v_next_major CONSTANT INTEGER := 3;
  v_next_minor CONSTANT INTEGER := 1;
  v_next_patch CONSTANT INTEGER := 0;
  
  -- verify if the 'Online Index Build' option is supported (disabled in Oracle XE)
  online_index_build V$OPTION.VALUE%TYPE := NULL;

--
-- Add index if it does not exist
PROCEDURE add_index(index_name  VARCHAR2
                   ,index_stmt  VARCHAR2)
IS
  index_exists EXCEPTION;                  --index already exists
  PRAGMA EXCEPTION_INIT(index_exists,-955); -- associate with error ORA-00955
BEGIN
  IF online_index_build IS NULL THEN
    -- verify if Online Index Build is enabled
    SELECT VALUE INTO online_index_build FROM V$OPTION WHERE PARAMETER='Online Index Build';
    IF online_index_build != 'TRUE' THEN
      DBMS_OUTPUT.PUT_LINE('WARNING: online index build is not supported');
    END IF;    
  END IF;
    
  IF online_index_build = 'TRUE' THEN
    EXECUTE IMMEDIATE ('CREATE INDEX ' || index_name || ' ON ' || index_stmt || ' ONLINE');
  ELSE
    EXECUTE IMMEDIATE ('CREATE INDEX ' || index_name || ' ON ' || index_stmt);
  END IF;    
  DBMS_OUTPUT.PUT_LINE('INFO: Index [' || index_name || '] added');
EXCEPTION
WHEN index_exists THEN
  DBMS_OUTPUT.PUT_LINE('INFO: Index [' || index_name || '] already exists. Skip it');
  RETURN;
END; --add_index
    
--
-- Drop Table procedure
PROCEDURE drop_if_exists(table_name VARCHAR2)
IS
  table_inexistent EXCEPTION;                   -- table does not exists
  PRAGMA EXCEPTION_INIT(table_inexistent,-942); -- associate with error ORA-00942
BEGIN 
  EXECUTE IMMEDIATE ('DROP TABLE '|| table_name || ' CASCADE CONSTRAINTS');
EXCEPTION
WHEN table_inexistent THEN
  RETURN;
END drop_if_exists;

--
-- Drop Sequence procedure
PROCEDURE drop_seq_if_exists(seq_name VARCHAR2)
IS
  seq_inexistent EXCEPTION;                  -- seq does not exists
  PRAGMA EXCEPTION_INIT(seq_inexistent,-2289);-- associate with error ORA-02289
BEGIN 
  EXECUTE IMMEDIATE ('DROP SEQUENCE '|| seq_name);
EXCEPTION
WHEN seq_inexistent THEN
  RETURN;
END drop_seq_if_exists;

--
-- Add a column to an existing table
PROCEDURE add_column(table_name  VARCHAR2
                    ,column_name VARCHAR2
                    ,column_stmt VARCHAR2)
IS
  column_exists EXCEPTION;                  --column already exists
  PRAGMA EXCEPTION_INIT(column_exists,-1430); -- associate with error ORA-01430
BEGIN
  EXECUTE IMMEDIATE ('ALTER TABLE ' || table_name || ' ADD ' || column_name || ' ' || column_stmt);
  DBMS_OUTPUT.PUT_LINE('INFO: Column [' || table_name || '.' || column_name || '] added');
EXCEPTION
WHEN column_exists THEN
  DBMS_OUTPUT.PUT_LINE('INFO: Column [' || table_name || '.' || column_name || '] already exists. Skip it');
  RETURN;
END; --add_column

--
-- Main
BEGIN

  DBMS_OUTPUT.PUT_LINE('Applying Patch...');
  -- Get schema version
  SELECT MAJOR,MINOR,PATCH INTO v_major,v_minor,v_patch FROM T_SCHEMA_VERS WHERE ROWNUM=1;
  -- check version
  IF v_major = v_req_major AND  v_minor = v_req_minor AND (v_patch BETWEEN  v_req_patch_min AND v_req_patch_max) THEN
      -- apply patch
      
      -- Add a schema state column that can be used to check if a previous upgrade attempt failed
      DBMS_OUTPUT.PUT_LINE('Upgrading T_SCHEMA_VERS...');
      add_column('t_schema_vers','state','VARCHAR2(24)');
      -- TODO: I couldn't find a way to query the new column since the EXECUTE IMMEDIATE is 
      --       performed after the script is compiled. Next upgrade script could make 
      --       use of this field to disply a warning in case a previous upgrade attempt 
      --       failed: in this case, the t_schema_vers.state would not be NULL.

      -- set schema state 
      EXECUTE IMMEDIATE('UPDATE T_SCHEMA_VERS SET STATE=''Upgrading''');
      DBMS_OUTPUT.PUT_LINE('Upgrading schema ' || v_major || '.' || v_minor || '.' || v_patch || ' to ' || v_next_major || '.' || v_next_minor || '.' || v_next_patch || ' ...');

      -- add t_site_group table
      DBMS_OUTPUT.PUT_LINE('Adding T_SITE_GROUP...');
      drop_if_exists('T_SITE_GROUP');
      EXECUTE IMMEDIATE('CREATE TABLE t_site_group (
                          group_name     VARCHAR2(100)
                                        CONSTRAINT site_group_gname_not_null NOT NULL
                        , site_name      VARCHAR2(100)
                                        CONSTRAINT site_group_sname_not_null NOT NULL
                        , priority       INTEGER DEFAULT 3
                        , CONSTRAINT sitegroup_pk PRIMARY KEY (group_name, site_name)
                       )');
      add_index('site_group_sname_id', 't_site_group(site_name)');
      add_index('site_group_gname_id', 't_site_group(group_name)');

      -- add t_channel properties
      DBMS_OUTPUT.PUT_LINE('Adding T_CHANNEL properties...');
      add_column('t_channel','blocksize'          ,'VARCHAR2(24)');
      add_column('t_channel','http_to'            ,'INTEGER');
      add_column('t_channel','tx_loglevel'        ,'VARCHAR2(12)');
      add_column('t_channel','urlcopy_put_to'     ,'INTEGER');
      add_column('t_channel','urlcopy_putdone_to' ,'INTEGER');
      add_column('t_channel','urlcopy_get_to'     ,'INTEGER');
      add_column('t_channel','urlcopy_getdone_to' ,'INTEGER');
      add_column('t_channel','urlcopy_tx_to'      ,'INTEGER');
      add_column('t_channel','urlcopy_txmarks_to' ,'INTEGER');
      add_column('t_channel','srmcopy_direction'  ,'VARCHAR2(4)');
      add_column('t_channel','srmcopy_to'         ,'INTEGER');
      add_column('t_channel','srmcopy_refresh_to' ,'INTEGER');

      -- add t_job properties
      DBMS_OUTPUT.PUT_LINE('Adding T_JOB properties...');
      add_column('t_job','job_finished' ,'TIMESTAMP DEFAULT NULL');
      
      -- add t_file properties
      DBMS_OUTPUT.PUT_LINE('Adding T_FILE properties...');
      add_column('t_file','job_finished' ,'TIMESTAMP DEFAULT NULL');
      add_column('t_file','error_scope'  ,'VARCHAR2(32)');
      add_column('t_file','error_phase'  ,'VARCHAR2(32)');
      
      -- add t_transfer properties
      DBMS_OUTPUT.PUT_LINE('Adding T_TRANSFER properties...');
      add_column('t_transfer','job_finished' ,'TIMESTAMP DEFAULT NULL');
      add_column('t_transfer','error_scope'  ,'VARCHAR2(32)');
      add_column('t_transfer','error_phase'  ,'VARCHAR2(32)');
      add_column('t_transfer','finish_time'  ,'TIMESTAMP WITH TIME ZONE');
            
      -- add t_stage_req properties
      DBMS_OUTPUT.PUT_LINE('Adding T_STAGE_REQ properties...');
      add_column('t_stage_req','job_finished' ,'TIMESTAMP DEFAULT NULL');
      add_column('t_stage_req','error_scope'  ,'VARCHAR2(32)');
      add_column('t_stage_req','error_phase'  ,'VARCHAR2(32)');
      add_column('t_stage_req','finish_time'  ,'TIMESTAMP WITH TIME ZONE');

      -- add indexes
      DBMS_OUTPUT.PUT_LINE('Adding T_JOB indices...');
      add_index('job_jobfinished_id'     , 't_job(job_finished)');
      DBMS_OUTPUT.PUT_LINE('Adding T_FILE indices...');
      add_index('file_jobfinished_id'    , 't_file(job_finished)');
      DBMS_OUTPUT.PUT_LINE('Adding T_TRANSFER indices...');
      add_index('transfer_jobfinished_id', 't_transfer(job_finished)');
      add_index('transfer_finishtime_id' , 't_transfer(finish_time)');
      DBMS_OUTPUT.PUT_LINE('Adding T_STAGE_REQ indices...');
      add_index('stagereq_jobfinished_id', 't_stage_req(job_finished)');
      add_index('stagereq_finishtime_id' , 't_stage_req(finish_time)');
      -- TODO: Add new indices for optimization

      -- add trigger to set teh job_finished time in all the tables. It will be used for partitioning
      DBMS_OUTPUT.PUT_LINE('Adding SET_JOB_FINISHED trigger...');
      EXECUTE IMMEDIATE('CREATE OR REPLACE TRIGGER SET_JOB_FINISHED
        BEFORE UPDATE ON T_JOB
        FOR EACH ROW 
        WHEN ( new.job_state IN (''Finished'',''FinishedDirty'',''Failed'',''Canceled'') ) 
        BEGIN
         SELECT SYSTIMESTAMP INTO :new.job_finished FROM DUAL;
         UPDATE T_FILE SET job_finished=:new.job_finished WHERE T_FILE.job_id=:new.job_id;
         UPDATE T_TRANSFER SET job_finished=:new.job_finished WHERE T_TRANSFER.job_id=:new.job_id;
         UPDATE T_STAGE_REQ SET job_finished=:new.job_finished WHERE T_STAGE_REQ.job_id=:new.job_id;
        END;');
      -- TODO: Set the job_finished time on the existing and already completed jobs 

      DBMS_OUTPUT.PUT_LINE('Updating schema version...');
      UPDATE T_SCHEMA_VERS SET MAJOR=v_next_major, MINOR=v_next_minor, PATCH=v_next_patch; 
      EXECUTE IMMEDIATE ('UPDATE T_SCHEMA_VERS SET STATE=NULL'); 
      DBMS_OUTPUT.PUT_LINE('SUCCESS: Patch applied. New version is ' || TO_CHAR(v_next_major) || '.' || TO_CHAR(v_next_minor) || '.' || TO_CHAR(v_next_patch) );
  ELSIF v_major = v_next_major AND  v_minor = v_next_minor AND v_patch = v_next_patch THEN
     DBMS_OUTPUT.PUT_LINE('SUCCESS: Patch already applied. Version is ' || TO_CHAR(v_next_major) || '.' || TO_CHAR(v_next_minor) || '.' || TO_CHAR(v_next_patch) );
  ELSE
     DBMS_OUTPUT.PUT_LINE('ERROR: Cannot apply patch. Version ' || TO_CHAR(v_major) || '.' || TO_CHAR(v_minor) || '.' || TO_CHAR(v_patch) || ' does not match the required one (' || TO_CHAR(v_req_major) || '.' || TO_CHAR(v_req_minor) || '.' || TO_CHAR(v_req_patch_min) || '-' ||  TO_CHAR(v_req_patch_max) || ')');
  END IF;
END;
/
